// *** 7 Colors for Windows CE ***
// Version 1.4
// Franois PESSAUX June 2001
// This program is freeware
// Use its sources as you want provided that you leave this copyright unchanged
// and that you drop a small word telling that I'm the author of these sources.
// Thank's a lot
//   --Francois
#include "stdafx.h"



#define FREE_VSPACE 2
#define FREE_HSPACE 2

#define STANDARD_FLAG 0
#define NO_START_SPACE_FLAG 1
#define NO_STOP_SPACE_FLAG 2



void vertical_fence_at (struct game *game, int column, int flag)
{
	struct cell *board ;
	int start_line, stop_line, line, cell, count=0 ;

	board = game->board ;

	if (flag & NO_START_SPACE_FLAG) start_line = 0 ;
	else start_line = FREE_VSPACE ;
	if (flag & NO_STOP_SPACE_FLAG) stop_line = board_height ;
	else stop_line = board_height - FREE_VSPACE ;

	// 2 * board_width because there are indeed twices more line than board_height, remember !!!
	cell = column + start_line * 2 * board_width ;

	for (line=start_line; line<stop_line; line++) {
		if (board[cell].owner != OWNER_FORBIDDEN) {
			board[cell].color = COLOR_BLACK ;
			board[cell].owner = OWNER_FORBIDDEN ;
			count++ ;
			}
		// 2 * board_width because there are indeed twices more lines than board_height, remember !!!
		cell += (2 * board_width) ;
		}
	game->number_forbidden_cells += count ;
}



void horizontal_fence_at (struct game *game, int line, int flag)
{
	struct cell *board ;
	int start_col, stop_col, col, cell, count=0 ;

	board = game->board ;

	// board_width * 2 because there are indeed twices more lines than board_height, remember !!!
	if (flag & NO_START_SPACE_FLAG) start_col = 0 ;
	else start_col = FREE_HSPACE ;
	if (flag & NO_STOP_SPACE_FLAG) stop_col = board_width ;
	else stop_col = board_width - FREE_HSPACE ;

	cell = line * (board_width * 2) + start_col ;
	for (col=start_col; col<stop_col; col++) {
		if (board[cell].owner != OWNER_FORBIDDEN) {
			board[cell].color = COLOR_BLACK ;
			board[cell].owner = OWNER_FORBIDDEN ;
			count++ ;
			}
		cell++ ;;
		}
	game->number_forbidden_cells += count ;
}



void one_vertical_fence (struct game *game)
{
	if (board_width>4) vertical_fence_at (game, (board_width / 2), STANDARD_FLAG) ;
}


void two_vertical_fences (struct game *game)
{
	int tmp, comp_top, comp_bottom ;

	if (board_width>4) {
		tmp = board_width / 3 ;
		if ((board_width % 3) == 0) { comp_top = 1 ; comp_bottom = 0 ; }
		else { comp_top = 0 ; comp_bottom = 1 ; }
		vertical_fence_at (game, tmp - (1 * comp_top), STANDARD_FLAG) ;
		vertical_fence_at (game, board_width - (1 * comp_bottom) - tmp, STANDARD_FLAG) ;
		}
}


void two_vertical_fences_no_space (struct game *game)
{
	int tmp, comp_top, comp_bottom ;

	tmp = board_width / 3 ;
	if ((board_width % 3) == 0) { comp_top = 1 ; comp_bottom = 0 ; }
	else { comp_top = 0 ; comp_bottom = 1 ; }
	if (board_width>4) {
		vertical_fence_at (game, tmp - (1 * comp_top), NO_STOP_SPACE_FLAG) ;
		vertical_fence_at (game, board_width - (1 * comp_bottom) - tmp, NO_START_SPACE_FLAG) ;
	}
}


void one_horizontal_fence (struct game *game)
{
	if (board_height>4) horizontal_fence_at (game, (board_height / 2), STANDARD_FLAG) ;
}


void two_horizontal_fences (struct game *game)
{
	int tmp, comp_left, comp_right ;
	
	if (board_height>4) {
		tmp = board_height / 3 ;
		if ((board_height % 3) == 0) { comp_left = 1 ; comp_right = 0 ; }
		else { comp_left = 0 ; comp_right = 1 ; }
		horizontal_fence_at (game, tmp - (1 * comp_left), STANDARD_FLAG) ;
		horizontal_fence_at (game, board_height - (1 * comp_right) - tmp, STANDARD_FLAG) ;
	}
}


void two_horizontal_fences_no_space (struct game *game)
{
	int tmp, comp_left, comp_right ;
	
	if (board_height>4) {
		tmp = board_height / 3 ;
		if ((board_height % 3) == 0) { comp_left = 1 ; comp_right = 0 ; }
		else { comp_left = 0 ; comp_right = 1 ; }
		horizontal_fence_at (game, tmp - (1 * comp_left), NO_STOP_SPACE_FLAG) ;
		horizontal_fence_at (game, board_height - (1 * comp_right) - tmp, NO_START_SPACE_FLAG) ;
		}
}


int north_east (int i) {
  int lnum ;

  if ((i < board_width) || ((i % (2 * board_width)) == (board_width - 1)))
    return (OUT_OF_BOARD) ;
  lnum = i / board_width ;
  if ((lnum % 2) == 0) return (i - (board_width - 1)) ;
  else return (i - board_width) ;
}


int north_west (int i) {
  int lnum ;

  if ((i < board_width) || ((i % (2 * board_width)) == board_width))
    return (OUT_OF_BOARD) ;
  lnum = i / board_width ;
  if ((lnum % 2) == 0) return (i - board_width) ;
  else return (i - (board_width + 1)) ;
}


int south_east (int i) {
  int lnum ;

  if ((i >= (board_width * board_height * 2 - board_width)) ||
      ((i % (2 * board_width)) == (board_width - 1)))
    return (OUT_OF_BOARD) ;
  lnum = i / board_width ;
  if ((lnum % 2) == 0) return (i + (board_width + 1)) ;
  else return (i + board_width) ;
}


int south_west (int i) {
  int lnum ;

  if ((i >= (board_width * board_height * 2 - board_width)) ||
      ((i % (2 * board_width)) == board_width))
    return (OUT_OF_BOARD) ;
  lnum = i / board_width ;
  if ((lnum % 2) == 0) return (i + board_width) ;
  else return (i + (board_width - 1)) ;
}



int fill_connected (struct game *game, int player_num, int c, int with_color) {
  struct cell *board ;
  int tmp, count ;

  board = game->board ;
  count = 0 ;

	// Stop test: the case has already be seen.
	if (board[c].marker & SEEN_MARKER) return (0) ;
	board[c].marker |= SEEN_MARKER ;

  switch (board[c].owner) {
  case OWNER_NOBODY:
    if (board[c].color == with_color) {
      // Acquire this new case. No need to change the case's color : it's already good.
      count++ ;
      board[c].owner = player_num ;
			// Set the marker
			board[c].marker |= DONTFILL_MARKER ;
      tmp = north_east (c) ;
      if (tmp != OUT_OF_BOARD)
				count += fill_connected (game, player_num, tmp, with_color) ;
      tmp = north_west (c) ;
      if (tmp != OUT_OF_BOARD)
				count += fill_connected (game, player_num, tmp, with_color) ;
      tmp = south_east (c) ;
      if (tmp != OUT_OF_BOARD)
				count += fill_connected (game, player_num, tmp, with_color) ;
      tmp = south_west (c) ;
      if (tmp != OUT_OF_BOARD)
				count += fill_connected (game, player_num, tmp, with_color) ;
		}
		break ;														// End of OWNER_NOBODY

  case OWNER_PLAYER0:
  case OWNER_PLAYER1:
    if (board[c].owner == player_num) {
      // Just change color of the board, and find adjacent cases.
      count++ ;
      board[c].color = with_color ;
			// Set the marker
			board[c].marker |= DONTFILL_MARKER ;
      tmp = north_east (c) ;
      if (tmp != OUT_OF_BOARD)
				count += fill_connected (game, player_num, tmp, with_color) ;
      tmp = north_west (c) ;
      if (tmp != OUT_OF_BOARD)
				count += fill_connected (game, player_num, tmp, with_color) ;
      tmp = south_east (c) ;
      if (tmp != OUT_OF_BOARD)
				count += fill_connected (game, player_num, tmp, with_color) ;
      tmp = south_west (c) ;
      if (tmp != OUT_OF_BOARD)
				count += fill_connected (game, player_num, tmp, with_color) ;
		}
		break ;														// End of OWNER_PLAYERx

	case OWNER_FORBIDDEN:	
		board[c].marker |= DONTFILL_MARKER ;
		break ;														// End of OWNER_FORBIDDEN:	
  }

  return (count) ;
}



// Notes:
// Cell "cell" is already colored.
// Integer "remaining" is the number of neighbour cells of "cell" that one must
// color from the current cell "cell".
int try_spread (struct cell *board, int color, int cell, int remaining)
{
	int tmp, i ;
	int direction = 0 ;
	int filled = 0 ;

	direction = rand () % 4 ;

	for (i=0; i<4; i++) {
		if (filled == remaining) break ;
		direction = (direction + 1) % 4 ;

		switch (direction) {
			case 0: tmp = north_east (cell) ; break ;
			case 1: tmp = north_west (cell) ; break ;
			case 2: tmp = south_east (cell) ; break ;
			case 3: tmp = south_west (cell) ; break ;
			}																		// End of switch (direction)

		if ((tmp != OUT_OF_BOARD) && (board[tmp].color == COLOR_BLACK)) {
			board[tmp].color = color ;
			filled = filled + 1 + try_spread (board, color, tmp, remaining - 1) ;
			}
		}

	return (filled) ;
}



struct game* alloc_game () {
  struct cell *board ;
  struct game *game ;
  int i, count ;
  int board_array_size ;
  int color, player0_color, player1_color ;

  players[OWNER_PLAYER0].pstart = 0 ;
  players[OWNER_PLAYER0].pcolor = -1 ;
  players[OWNER_PLAYER0].pcount = 1 ;

  players[OWNER_PLAYER1].pstart = 0 ;
  players[OWNER_PLAYER1].pcolor = -1 ;
  players[OWNER_PLAYER1].pcount = 1 ;

  board_array_size = board_width * board_height * 2 ;
  board = (struct cell*) malloc (board_array_size * sizeof (struct cell)) ;

	// First, set the color to "nothing", set the owner to "nobody", and clear the initial marker
	for (i=0; i<board_array_size; i++) {
		board[i].color = COLOR_BLACK ;
		board[i].owner = OWNER_NOBODY ;
		board[i].marker = EMPTY_MARKER ;
		}

  for (i=0; i<board_array_size; i++) {
		// Only change this case's color if it has not be set before.
		if (board[i].color == COLOR_BLACK) {
			color = (rand () % 7) ;
			board[i].color = color ;
			// Bias neighbour colors to introduce game difficulty.
			try_spread (board, color, i, spreading_factor - 1) ;
			}
		}

  // Set initial players information
  // Initial position : lower left corner
  players[OWNER_PLAYER0].pstart = (board_width * board_height * 2) - board_width ;
  // Initial position : upper right corner
  players[OWNER_PLAYER1].pstart = board_width - 1 ;

  player0_color = board[players[OWNER_PLAYER0].pstart].color ;
	players[OWNER_PLAYER0].pcolor = player0_color ;
	board[(players[OWNER_PLAYER0]).pstart].owner = OWNER_PLAYER0 ;

	// Avoid the 2 players to have the same starting color.
  player1_color = board[players[OWNER_PLAYER1].pstart].color ;
	if (player1_color == player0_color)	player1_color = (player1_color++) % 7 ;
	players[OWNER_PLAYER1].pcolor = player1_color ;
	board[(players[OWNER_PLAYER1]).pstart].owner = OWNER_PLAYER1 ;

	// Create the game structure
  game = (struct game*) malloc (sizeof (struct game)) ;
  game->board_array_size = board_array_size ;
	game->number_forbidden_cells = 0 ;
  game->board = board ;

	if (fence_style & FENCE_STYLE_1_VERTICAL) one_vertical_fence (game) ;
	if (fence_style & FENCE_STYLE_2_VERTICAL) two_vertical_fences (game) ;
	if (fence_style & FENCE_STYLE_1_HORIZONTAL) one_horizontal_fence (game) ;
	if (fence_style & FENCE_STYLE_2_HORIZONTAL) two_horizontal_fences (game) ;
	if (fence_style & FENCE_STYLE_2_VERTICAL_NO_SPACE) two_vertical_fences_no_space (game) ;
	if (fence_style & FENCE_STYLE_2_HORIZONTAL_NO_SPACE) two_horizontal_fences_no_space (game) ;

	// Play one empty turn for each player to acquire cells of their initial color
	count = fill (game, OWNER_PLAYER0, players[OWNER_PLAYER0].pstart, player0_color) ;
  players[OWNER_PLAYER0].pcount = count ;
	count = fill (game, OWNER_PLAYER1, players[OWNER_PLAYER1].pstart, player1_color) ;
  players[OWNER_PLAYER1].pcount = count ;

	// Finally return the game ready to use
	turn = OWNER_PLAYER0 ;
  return (game) ;
}


void release_game (struct game *game) {

  if (game == NULL) return ;
  free (game->board) ;
  free (game) ;
	game = NULL ;
}


struct game* smart_copy_game (struct game *game) {
  struct cell *new_board ;
  struct game *new_game ;
  int board_array_size ;
  int i ;

  board_array_size = board_width * board_height * 2 ;
  new_board = (struct cell*) malloc (board_array_size * sizeof (struct cell)) ;
  for (i=0; i<board_array_size; i++) {
    new_board[i].color = game->board[i].color ;
    new_board[i].owner = game->board[i].owner ;
    new_board[i].marker = game->board[i].marker ;
  }
  new_game = (struct game*) malloc (sizeof (struct game)) ;
  new_game->board_array_size = game->board_array_size ;
  new_game->board = new_board ;
  return (new_game) ;
}



// Just set the marker on all reachable cases from the case c
// that does not belong to the player "owner".
// We are not allowed to "walk" on the fences !!!
void mark_reachable_from_and_not_belonging_to (struct game *game, int c,
					       int owner)
{
  int tmp ;
  struct cell *board ;

  board = game->board ;
	
	if (game->board[c].marker & SEEN_MARKER) return ;
	game->board[c].marker |= SEEN_MARKER ;

	if (game->board[c].owner != owner) {
		board[c].marker |= DONTFILL_MARKER ;
		// We are not allowed to jump the fences !
		if (game->board[c].owner != OWNER_FORBIDDEN) {
			tmp = north_east (c) ;
			if (tmp != OUT_OF_BOARD)
				mark_reachable_from_and_not_belonging_to (game, tmp, owner) ;
			tmp = north_west (c) ;
			if (tmp != OUT_OF_BOARD)
				mark_reachable_from_and_not_belonging_to (game, tmp, owner) ;
			tmp = south_east (c) ;
			if (tmp != OUT_OF_BOARD)
				mark_reachable_from_and_not_belonging_to (game, tmp, owner) ;
			tmp = south_west (c) ;
			if (tmp != OUT_OF_BOARD)
				mark_reachable_from_and_not_belonging_to (game, tmp, owner) ;
			}
		}
}



int fill_unmarked_cases (struct game *game,
					       int with_color, int owner_num) {
  int c ;
  int count = 0 ;
  struct cell *board ;
  struct player *player ;

  board = game->board ;
  player = &players[owner_num] ;
  for (c=0; c<game->board_array_size; c++) {
    if (!(board[c].marker & DONTFILL_MARKER)) {
      board[c].color = with_color ;
      board[c].owner = owner_num ;
      count++ ;
      player->pcount++ ;
    }
  }
  return (count) ;
}



int fill (struct game *game, int player_num, int current_case, int with_color) {
  int tmp0, tmp1, from_c, c ;
  struct cell *board ;

  // Fill all connected cases with with_color
  tmp0 = fill_connected (game, player_num, current_case, with_color) ;

	// Reset the markers telling that we already visited a case
	board = game->board ;
	for (c=0; c<game->board_array_size; c++) board[c].marker &= (~SEEN_MARKER) ;

  // Determine the opponent starting location
  if (player_num == OWNER_PLAYER0) from_c = players[OWNER_PLAYER1].pstart ;
  else from_c = players[OWNER_PLAYER0].pstart ;
  
	// Mark all cases not belonging to the current player and that reachable from
  // the opponent starting position.
  mark_reachable_from_and_not_belonging_to (game, from_c, player_num) ;
  
	// Now fill the cases remaining that are non marked with DONTFILL_MARKER
  tmp1 = fill_unmarked_cases (game, with_color,	player_num) ;

	// Now clean the markers
	for (c=0; c<game->board_array_size; c++) board[c].marker = EMPTY_MARKER ;
  return (tmp0 + tmp1) ;
}



int computer_ai (struct game *game) {
  int authorized_colors[5] ;
  struct game *tmp_game ;
  int old_pcount, new_pcount ;
  int i,j, color, best_color ;

  for (i=0, j=0; i<7; i++) {
    if ((i != players[OWNER_PLAYER0].pcolor) &&	(i != players[OWNER_PLAYER1].pcolor)) {
      authorized_colors[j] = i ;
      j++ ;
    }
  }

  old_pcount = 0 ;
  best_color = authorized_colors[0] ;
  for (i=0; i<5; i++) {
    tmp_game = smart_copy_game (game) ;
    color = authorized_colors[i] ;
    new_pcount = fill (tmp_game, computer_id, players[computer_id].pstart, color) ;
    if (new_pcount > old_pcount) {
      old_pcount = new_pcount ;
      best_color = color ;
    }
    release_game (tmp_game) ;
  }
  return (best_color) ;
}



int check_end_of_game (struct game *game) {
  double player0_percent, player1_percent ;

  player0_percent =
    ((double) (players[OWNER_PLAYER0].pcount)) /
    (current_game->board_array_size - current_game->number_forbidden_cells) * 100.0 ;
  player1_percent =
    ((double) (players[OWNER_PLAYER1].pcount)) /
    (current_game->board_array_size - current_game->number_forbidden_cells) * 100.0 ;
  if (player0_percent >= 50.0) return (OWNER_PLAYER0) ;
  if (player1_percent >= 50.0) return (OWNER_PLAYER1) ;
  return (OWNER_NOBODY) ;
}
