// *** 7 Colors for Windows CE ***
// Version 1.2
// Franois PESSAUX November 2000
// This program is freeware
// Use its sources as you whant provided that you leave this copyright unchanged
// and that you drop a small word telling that I'm the author of these sources.
// Thank's a lot
//   --Francois
#include "stdafx.h"



int north_east (int i) {
  int lnum ;

  if ((i < board_width) || ((i % (2 * board_width)) == (board_width - 1)))
    return (OUT_OF_BOARD) ;
  lnum = i / board_width ;
  if ((lnum % 2) == 0) return (i - (board_width - 1)) ;
  else return (i - board_width) ;
}


int north_west (int i) {
  int lnum ;

  if ((i < board_width) || ((i % (2 * board_width)) == board_width))
    return (OUT_OF_BOARD) ;
  lnum = i / board_width ;
  if ((lnum % 2) == 0) return (i - board_width) ;
  else return (i - (board_width + 1)) ;
}


int south_east (int i) {
  int lnum ;

  if ((i >= (board_width * board_height * 2 - board_width)) ||
      ((i % (2 * board_width)) == (board_width - 1)))
    return (OUT_OF_BOARD) ;
  lnum = i / board_width ;
  if ((lnum % 2) == 0) return (i + (board_width + 1)) ;
  else return (i + board_width) ;
}


int south_west (int i) {
  int lnum ;

  if ((i >= (board_width * board_height * 2 - board_width)) ||
      ((i % (2 * board_width)) == board_width))
    return (OUT_OF_BOARD) ;
  lnum = i / board_width ;
  if ((lnum % 2) == 0) return (i + board_width) ;
  else return (i + (board_width - 1)) ;
}



int fill_connected (struct game *game, int player_num, int c, int with_color) {
  struct cell *board ;
  int old_player_color ;
  int tmp, count ;

  board = game->board ;
  old_player_color = players[player_num].pcolor ;
  count = 0 ;

  switch (board[c].owner) {
  case OWNER_NOBODY:
    if (board[c].color == with_color) {
      // Acquire this new case. No need to change the case's color : it's already good.
      count++ ;
      board[c].owner = player_num ;
      // Set the marker
      board[c].marker = TRUE ;
      tmp = north_east (c) ;
      if (tmp != OUT_OF_BOARD)
				count += fill_connected (game, player_num, tmp, with_color) ;
      tmp = north_west (c) ;
      if (tmp != OUT_OF_BOARD)
				count += fill_connected (game, player_num, tmp, with_color) ;
      tmp = south_east (c) ;
      if (tmp != OUT_OF_BOARD)
				count += fill_connected (game, player_num, tmp, with_color) ;
      tmp = south_west (c) ;
      if (tmp != OUT_OF_BOARD)
				count += fill_connected (game, player_num, tmp, with_color) ;
		}
		break ;														// End of OWNER_NOBODY

  case OWNER_PLAYER0:
  case OWNER_PLAYER1:
    if ((board[c].owner == player_num) &&
			  (board[c].color == old_player_color)) {
      // Just change color of the board, and find adjacent cases.
      count++ ;
      board[c].color = with_color ;
      // Set the marker
      board[c].marker = TRUE ;
      tmp = north_east (c) ;
      if (tmp != OUT_OF_BOARD)
				count += fill_connected (game, player_num, tmp, with_color) ;
      tmp = north_west (c) ;
      if (tmp != OUT_OF_BOARD)
				count += fill_connected (game, player_num, tmp, with_color) ;
      tmp = south_east (c) ;
      if (tmp != OUT_OF_BOARD)
				count += fill_connected (game, player_num, tmp, with_color) ;
      tmp = south_west (c) ;
      if (tmp != OUT_OF_BOARD)
				count += fill_connected (game, player_num, tmp, with_color) ;
		}
		break ;														// End of OWNER_PLAYERx
  }

  return (count) ;
}



struct game* alloc_game () {
  struct cell *board ;
  struct game *game ;
  int i, c ;
  int board_array_size ;
  int color, excluded_color0, excluded_color1, player0_color, player1_color ;
  int rand_color_tab[6] ;

  players[OWNER_PLAYER0].pstart = 0 ;
  players[OWNER_PLAYER0].pcolor = -1 ;
  players[OWNER_PLAYER0].pcount = 1 ;

  players[OWNER_PLAYER1].pstart = 0 ;
  players[OWNER_PLAYER1].pcolor = -1 ;
  players[OWNER_PLAYER1].pcount = 1 ;

  board_array_size = board_width * board_height * 2 ;
  board = (struct cell*) malloc (board_array_size * sizeof (struct cell)) ;
  for (i=0; i<board_array_size; i++) {
    color = (rand () % 7) ;
    board[i].color = color ;
    board[i].owner = OWNER_NOBODY ;
    board[i].marker = FALSE ;
  }
  // Set initial players position
  // Lower left corner
  players[OWNER_PLAYER0].pstart =
    (board_width * board_height * 2) - board_width ;
  // Upper right corner
  players[OWNER_PLAYER1].pstart = board_width - 1 ;
  // Set color of the board at the initial players positions to avoid
  // one player to have several cases at the beginning of the game.
  excluded_color0 =
    board[players[OWNER_PLAYER0].pstart - board_width + 1].color ;
  excluded_color1 =
    board[players[OWNER_PLAYER1].pstart + board_width - 1].color ;
  for (c=0, i=0; c<7; c++) {
    if ((c != excluded_color0) && (c != excluded_color1)) {
      rand_color_tab[i] = c ;
      i++ ;
    }
  }
  player0_color = rand_color_tab[(rand () %  i)] ;
  for (c=0, i=0; c<7; c++) {
    if ((c != excluded_color0) && (c != excluded_color1) &&
	(c != player0_color)) {
      rand_color_tab[i] = c ;
      i++ ;
    }
  }
  player1_color = rand_color_tab[(rand () %  i)] ;
  // Update the players' structure.
  players[OWNER_PLAYER0].pcolor = player0_color ;
  board[(players[OWNER_PLAYER0]).pstart].owner = OWNER_PLAYER0 ;
  board[(players[OWNER_PLAYER0]).pstart].color = player0_color ;
  players[OWNER_PLAYER1].pcolor = player1_color ;
  board[(players[OWNER_PLAYER1]).pstart].owner = OWNER_PLAYER1 ;
  board[(players[OWNER_PLAYER1]).pstart].color = player1_color ;
  turn = OWNER_PLAYER0 ;
  // Finally, return the game ready to use.
  game = (struct game*) malloc (sizeof (struct game)) ;
  game->board_array_size = board_array_size ;
  game->board = board ;
  return (game) ;
}


void release_game (struct game *game) {

  if (game == NULL) return ;
  free (game->board) ;
  free (game) ;
}


struct game* smart_copy_game (struct game *game) {
  struct cell *new_board ;
  struct game *new_game ;
  int board_array_size ;
  int i ;

  board_array_size = board_width * board_height * 2 ;
  new_board = (struct cell*) malloc (board_array_size * sizeof (struct cell)) ;
  for (i=0; i<board_array_size; i++) {
    new_board[i].color = game->board[i].color ;
    new_board[i].owner = game->board[i].owner ;
    new_board[i].marker = game->board[i].marker ;
  }
  new_game = (struct game*) malloc (sizeof (struct game)) ;
  new_game->board_array_size = game->board_array_size ;
  new_game->board = new_board ;
  return (new_game) ;
}



// Just set the marker on all reachable cases from the case c
void mark_reachable_from_and_not_belonging_to (struct game *game, int c,
					       int owner) {
  int tmp ;
  struct cell *board ;

  board = game->board ;
  if (! game->board[c].marker) {
    if (game->board[c].owner != owner) {
      board[c].marker = TRUE ;
      tmp = north_east (c) ;
      if (tmp != OUT_OF_BOARD)
				mark_reachable_from_and_not_belonging_to (game, tmp, owner) ;
      tmp = north_west (c) ;
      if (tmp != OUT_OF_BOARD)
				mark_reachable_from_and_not_belonging_to (game, tmp, owner) ;
      tmp = south_east (c) ;
      if (tmp != OUT_OF_BOARD)
				mark_reachable_from_and_not_belonging_to (game, tmp, owner) ;
      tmp = south_west (c) ;
      if (tmp != OUT_OF_BOARD)
				mark_reachable_from_and_not_belonging_to (game, tmp, owner) ;
    }
  }
}



int fill_unmarked_cases_and_and_reset_markers (struct game *game,
					       int with_color, int owner_num) {
  int c ;
  int count = 0 ;
  struct cell *board ;
  struct player *player ;

  board = game->board ;
  player = &players[owner_num] ;
  for (c=0; c<game->board_array_size; c++) {
    if (! board[c].marker) {
      board[c].color = with_color ;
      board[c].owner = owner_num ;
      count++ ;
      player->pcount++ ;
    }
    else
      board[c].marker = FALSE ;
  }
  return (count) ;
}



int fill (struct game *game, int player_num, int current_case, int with_color) {
  int tmp0, tmp1, from_c ;

  // Fill all connected cases with with_color
  tmp0 = fill_connected (game, player_num, current_case, with_color) ;
  // Determine the opponent starting location
  if (player_num == OWNER_PLAYER0)
    from_c = players[OWNER_PLAYER1].pstart ;
  else from_c = players[OWNER_PLAYER0].pstart ;
  // Mark all cases not belonging to the current player and that reachable from
  // the opponent starting position.
  mark_reachable_from_and_not_belonging_to (game, from_c, player_num) ;
  // Now fill the cases remaining that are non marked.
  tmp1 = fill_unmarked_cases_and_and_reset_markers (game, with_color,
		player_num) ;
  return (tmp0 + tmp1) ;
}



int computer_ai (struct game *game) {
  int authorized_colors[5] ;
  struct game *tmp_game ;
  int old_pcount, new_pcount ;
  int i,j, color, best_color ;

  for (i=0, j=0; i<7; i++) {
    if ((i != players[OWNER_PLAYER0].pcolor) &&
	(i != players[OWNER_PLAYER1].pcolor)) {
      authorized_colors[j] = i ;
      j++ ;
    }
  }
  old_pcount = players[computer_id].pcount ;
  best_color = authorized_colors[0] ;
  for (i=0; i<5; i++) {
    tmp_game = smart_copy_game (game) ;
    color = authorized_colors[i] ;
    new_pcount =
      fill (tmp_game, computer_id, players[computer_id].pstart, color) ;
    if (new_pcount > old_pcount) {
      old_pcount = new_pcount ;
      best_color = color ;
    }
    release_game (tmp_game) ;
  }
  return (best_color) ;
}



int check_end_of_game (struct game *game) {
  double player0_percent, player1_percent ;

  player0_percent =
    ((double) (players[OWNER_PLAYER0].pcount)) /
    current_game->board_array_size * 100.0 ;
  player1_percent =
    ((double) (players[OWNER_PLAYER1].pcount)) /
    current_game->board_array_size * 100.0 ;
  if (player0_percent >= 50.0) return (OWNER_PLAYER0) ;
  if (player1_percent >= 50.0) return (OWNER_PLAYER1) ;
  return (OWNER_NOBODY) ;
}
